#include <Arduino.h>
#include <SoftwareSerial.h>

#include <SentecSensors.h>
/***************************************
 *          RS485 SENSOR READOUT
 ****************************************/

SentecSensorRS485::SentecSensorRS485(SoftwareSerial *ser, byte add)
{
    address = add;
    RS485 = ser;
}

void SentecSensorRS485::write(byte queryFrame[], int length)
{
    // sends a message (bytes) to the sensor

    // Initialize the transmitter
    digitalWrite(serialCommunicationControlPin, HIGH);
    // Send message: request a reading from the sensor
    RS485->write(queryFrame, length);
    RS485->flush();
    // Initialize the receiver
    digitalWrite(serialCommunicationControlPin, LOW);
}

String SentecSensorRS485::getValueStr(float value)
{
    if (valid)
    {
        return String(value, 1);
    }
    else
    {
        return String("null");
    }
}

String SentecSensorRS485::getValueStr(int value)
{
    if (valid)
    {
        return String(value);
    }
    else
    {
        return String("null");
    }
}

void SentecSensorRS485::queryAddress()
{
    // request the address of the sensor with ONLY ONE SENSOR ON THE BUS

    byte tmp_addr = address; // store the address in a temporary byte
    address = 0xFF;          // change the address to FF (0) for address check
    readRegister(word(0x07, 0xD0), 2);
    address = tmp_addr; // set the original address back
}

void SentecSensorRS485::readRegister(int registerStartAddress)
{
    readRegister(registerStartAddress, 1);
}

void SentecSensorRS485::readRegister(int registerStartAddress, int registerLength)
{
    // function code 0x03: get data measured by the sensor
    byte query[8];
    query[0] = address;
    // function code
    query[1] = 0x03;
    // register start address
    query[2] = registerStartAddress >> 8;
    query[3] = registerStartAddress & 0xFF;
    // register length
    query[4] = registerLength >> 8;
    query[5] = registerLength & 0xFF;
    // calculate last two bytes (CRC check)
    calculateCRC(query, sizeof(query) - 2);
 //#   Serial.print("Query (get data): 0x");                    #Print bytes
 //#   printBytes(query, 8);
    // write the data request to the modbus line
    write(query, sizeof(query));
    // get response from sensor
    getResponse();
}

void SentecSensorRS485::writeRegister(int registerAddress, int value)
{
    // function code 0x06: change sensor settings
    // e.g. a new address, reset rainfal data...

    byte query[8];
    query[0] = address;
    // function code
    query[1] = 0x06;
    // register start address
    query[2] = registerAddress >> 8;
    query[3] = registerAddress & 0xFF;
    // register length
    query[4] = value >> 8;
    query[5] = value & 0xFF;
    calculateCRC(query, sizeof(query) - 2);
    Serial.print("Query (settings):    ");
    printBytes(query, 8);
    write(query, sizeof(query));
    getResponse();
}

void SentecSensorRS485::setAddress(byte add)
{
    // change the address of a sensor
    writeRegister(word(0x07, 0xD0), add);
    // TODO check response: matches the sent message exactly
    address = add;
}

void SentecSensorRS485::resetAnswerFrame()
{
    for (int i = 0; i < 10; i++)
    {
        answerFrame[i] = 0;
    }
}

bool SentecSensorRS485::getResponse()
{
    // reads the response of a sensor
    valid = true;
    int idx = 0;
    int byteReceived;
    // usual response length: changed in the while loop to match the response,
    //   changed only when reading data (then it's 7 or 9  bytes, sensor dpendent)
    int responseLength = 8;
    // reading an answer takes up to 39 milliseconds for 2 byte readRegister
    const int timeout = 200;
    const int retries = 3; //#editet to q
    size_t tries = 1;

    for (tries; tries <= retries; tries++)
    {
        // if we lose connection with the sensor, we get an array of zeros back
        resetAnswerFrame();

        unsigned long time = millis();
        while (idx < responseLength && (millis() - time) < timeout)
        {
            if (RS485->available())
            {
                byteReceived = RS485->read();
                // Serial.println(byteReceived, HEX);
                // check for first byte. It has to be the device address unless for broadcasts with address = 0xFF
                if (idx == 0 && address != 0xFF && byteReceived != address)
                {
                    Serial.print("Invalid byte. First byte needs to be address 0x");
                    Serial.print(address, HEX);
                    Serial.print(" but got 0x");
                    Serial.print(byteReceived, HEX);
                    Serial.println("instead");
                }
                else
                {
                    answerFrame[idx] = byteReceived;
                    // for reading register: third received byte is data length, read number of bytes accordingly
                    if (idx == 2 && answerFrame[1] == 0x03)
                    {
                        // 5 bytes for address, function, data length, CRC_H, CRC_L
                        responseLength = 5 + byteReceived;
                    }
                    idx++;
                }
            }
        }

        delay(10);
        Serial.print("Response: 0x");
        printBytes(answerFrame, responseLength);
        Serial.print("Tries: ");
        Serial.println(tries);
        Serial.print("Bytes received: ");
        Serial.println(idx);
        word crc_received = word(answerFrame[responseLength - 2], answerFrame[responseLength - 1]);
        word crc = calculateCRC(answerFrame, responseLength - 2);
        if (crc_received != word(crc))
        {
            Serial.print("CRC wrong: Expected ");
            printBytes(crc);
            Serial.print(" got ");
            printBytes(crc_received);
            Serial.println();
            valid = false;
            resetAnswerFrame();
        }

        if (answerFrame[0] == 0)
        {
            valid = false;
        }
        if(valid)
        {
            break;
        }
    }
    return valid;
}

unsigned int SentecSensorRS485::calculateCRC(byte query[], int length)
{
    // Change the last two bytes of the queryFrame to conform to a CRC check
    // Yes, this is necessary. No, I don't know exactly what it does.
    unsigned int tmp1, tmp2, flag;
    tmp1 = 0xFFFF;
    for (unsigned char i = 0; i < length; i++)
    {
        tmp1 = tmp1 ^ query[i];
        for (unsigned char j = 1; j <= 8; j++)
        {
            flag = tmp1 & 0x0001;
            tmp1 >>= 1;
            if (flag)
                tmp1 ^= 0xA001;
        }
    }
    // Reverse byte order.
    tmp2 = tmp1 >> 8;
    tmp1 = (tmp1 << 8) | tmp2;
    tmp1 &= 0xFFFF;
    // change last two query bytes
    query[length + 1] = tmp1;
    query[length] = tmp1 >> 8;

    return tmp1; // the returned value is already swapped - CRC_L byte is first & CRC_H byte is last
}

void SentecSensorRS485::printBytes(byte *data, int length)
{
    // prints 8-bit data in hex with leading zeroes
    char tmp[16];
    for (int i = 0; i < length; i++)
    {
        sprintf(tmp, "%.2X", data[i]);
        // sprintf(tmp, "0x%.2X",data[i]);
        Serial.print(tmp);
        Serial.print(" ");
    }
    Serial.println();
}

void SentecSensorRS485::printBytes(word data)
{
    char tmp[10];
    sprintf(tmp, "0x%.2X", data >> 8);
    Serial.print(tmp);
    sprintf(tmp, "%.2X", data & 0xFF);
    // sprintf(tmp, "0x%.2X",data[i]);
    Serial.print(tmp);
}

word SolarRadiationSensor::getSolarRadiation()
{
    readRegister(0, 1);
    glob = word(answerFrame[3], answerFrame[4]);
    Serial.print("Global solar radiation: ");
    Serial.print(glob, DEC);
    Serial.println(" W/m^2");
    return glob;
}

String SolarRadiationSensor::getSolarRadiationStr()
{
    return getValueStr((int)glob);
}

void RainGaugeSensor::queryTime()
{
    // get time setting of the rain gauge
    readRegister(0x34, 0x03);
}

void RainGaugeSensor::setTime()
{
    // set time of the rain gauge
    byte query[15];
    query[0] = address;
    query[1] = 0x10; // function code
    query[2] = 0x00; // register start address high
    query[3] = 0x34; // register start address low
    query[4] = 0x00; // register length high
    query[5] = 0x03; // register length low
    query[6] = 0x06; // data length
    // TODO parse timestamp... hard-code a random date for now.
    query[7] = 0x20;  // year
    query[8] = 0x04;  // month
    query[9] = 0x03;  // day
    query[10] = 0x17; // hours
    query[11] = 0x06; // minutes
    query[12] = 0x28; // seconds
    query[13] = 0x00; // CRC low
    query[14] = 0x00; // CRC high

    calculateCRC(query, sizeof(query) - 2);
    Serial.print("Query (set time):    ");
    printBytes(query, 15);
    write(query, sizeof(query));
    getResponse();
}

void RainGaugeSensor::resetPrecipitation()
{
    // clears rainfall rata from the rain gauge
    // delay resetting after the last register reading
    delay(100);
    Serial.println("Resetting precipitation sum");
    writeRegister(0x00, 0x5A);
    // TODO check response: matches the sent message exactly
}

void RainGaugeSensor::resetSensor()
{
    // no clue what the difference between this one and the previous one is...
    //   and the manual is not helpful, of course
    delay(100);
    Serial.println("Resetting precipitation sum");
    writeRegister(0x37, 0x03);
    // TODO check response: matches the sent message exactly
}

void RainGaugeSensor::getPrecipitation()
{
    // get all precipitation statistics
    readRegister(0, 0x0A);
    word prcpToday = word(answerFrame[3], answerFrame[4]);
    word prcpInstantaneous = word(answerFrame[5], answerFrame[6]); // TODO I'm pretty sure this value is bullshit (0.1 mm resolution)
    word prcpYesterday = word(answerFrame[7], answerFrame[8]);
    word prcpTotal = word(answerFrame[9], answerFrame[10]);
    word prcpHourly = word(answerFrame[11], answerFrame[12]);
    word prcpLastHourly = word(answerFrame[13], answerFrame[14]);
    word prcp24max = word(answerFrame[15], answerFrame[16]);
    word prcp24maxPeriod = word(answerFrame[17], answerFrame[18]);
    word prcp24min = word(answerFrame[19], answerFrame[20]);
    word prcp24minPeriod = word(answerFrame[21], answerFrame[22]);

    Serial.print("Today: ");
    Serial.print(prcpToday / 10.0, 1);
    Serial.println(" mm");
    Serial.print("Instantaneous: ");
    Serial.print(prcpInstantaneous / 10.0, 1);
    Serial.println(" mm");
    Serial.print("Total: ");
    Serial.print(prcpTotal / 10.0, 1);
    Serial.println(" mm");
    Serial.print("Hourly: ");
    Serial.print(prcpHourly / 10.0, 1);
    Serial.println(" mm");
    Serial.print("24 max: ");
    Serial.print(prcp24max / 10.0, 1);
    Serial.println(" mm");
    Serial.print("24 min: ");
    Serial.print(prcp24min / 10.0, 1);
    Serial.println(" mm");
}

word RainGaugeSensor::getInstantaneousPrecipitation()
{
    // gets tips of scale since the last reset, i.e. total precipitation (I THINK)
    //   manual says this is current precipitation - is it?
    readRegister(0, 0x01);
    precipitation = word(answerFrame[3], answerFrame[4]);
    Serial.print("Precipitation: ");
    Serial.print(precipitation / 10.0, 1);
    Serial.println(" mm");
    // resetPrecipitation();
    return precipitation;
}

String RainGaugeSensor::getPrecipitationStr()
{
    return getValueStr((float)(precipitation / 10.0));
}

word SoilMoistureSensor::getMoistureTemp()
{
    readRegister(0, 2); // start register at 0, read 2 variables (vwc, soil temp)
    moistureRaw = word(answerFrame[3], answerFrame[4]);
    if (answerFrame[5] < 0x80)
    {
        temperatureRaw = word(answerFrame[5], answerFrame[6]);
    }
    else
    {
        temperatureRaw = word(answerFrame[5], answerFrame[6]) - 65536;
    }
    Serial.print("Soil moisture: ");
    Serial.print((moistureRaw - moistureOffset) / 10.0, 1);
    Serial.println(" %");
    Serial.print("Temperature: ");
    Serial.print((temperatureRaw - temperatureOffset) / 10.0 , 1);
    Serial.println(" °C");
    return temperatureRaw;
}

float SoilMoistureSensor::getMoisture(){
   return (float)((moistureRaw - moistureOffset) / 10.0);
}

String SoilMoistureSensor::getMoistureStr()
{
    return getValueStr((float)((moistureRaw - moistureOffset) / 10.0));
}

String SoilMoistureSensor::getTemperatureStr()
{
    return getValueStr((float)((temperatureRaw - temperatureOffset) / 10.0));
}
